import aiger
from parsimonious import Grammar, NodeVisitor


PLTL_GRAMMAR = Grammar(u'''
phi =  since / or / and / implies / hist / past / vyest / neg
     / true / false / AP
or = "(" _ phi _ "|" _ phi _ ")"
implies = "(" _ phi _ "->" _ phi _ ")"
and = "(" _ phi _ "&" _ phi _ ")"
hist = "H" _ phi
past = "P" _ phi
vyest = "Z" _ phi
since = "[" _ phi _ "S" _ phi _ "]"
neg = "~" _ phi
true = "TRUE"
false = "FALSE"

_ = ~r" "*
AP = ~r"[a-zA-Z]" ~r"[a-zA-Z\\d]*"
EOL = "\\n"
''')


class PTLTLExpr(aiger.BoolExpr):
    def __call__(self, trc):
        if isinstance(trc, list):
            val, _ = self.aig.simulate(trc)[-1]
            return val[self.output]
        else:
            return aiger.BoolExpr.__call__(self, trc)

    def historically(self):
        return PTLTLExpr(self.aig >> hist_monitor(self.output))

    def once(self):
        return PTLTLExpr(self.aig >> past_monitor(self.output))

    def vyest(self):
        return PTLTLExpr(self.aig >> vyest_monitor(self.output))

    def since(self, other):
        monitor = since_monitor(self.output, other.output)
        return PTLTLExpr((self.aig | other.aig) >> monitor)


def atom(var):
    return PTLTLExpr(aiger.atom(var).aig)


# Secret Parsing based API for internal testing.


class PLTLVisitor(NodeVisitor):
    def generic_visit(self, _, children):
        return children

    def visit_phi(self, _, children):
        return children[0]

    def visit_AP(self, node, _):
        return PTLTLExpr(aiger.atom(node.text).aig)

    def visit_and(self, _, children):
        return children[2] & children[6]

    def visit_or(self, _, children):
        return children[2] | children[6]

    def visit_neg(self, _, children):
        return ~children[2]

    def visit_implies(self, _, children):
        return children[2].implies(children[6])

    def visit_vyest(self, _, children):
        return children[2].vyest()

    def visit_hist(self, _, children):
        return children[2].historically()

    def visit_past(self, _, children):
        return children[2].once()

    def visit_since(self, _, children):
        return children[2].since(children[6])

    def visit_true(self, *_):
        return PTLTLExpr(aiger.atom(True).aig)

    def visit_false(self, *_):
        return PTLTLExpr(aiger.atom(False).aig)


def vyest_monitor(name):
    return aiger.delay(
        inputs=[name],
        initials=[True],
        latches=[aiger.common._fresh()],
        outputs=[aiger.common._fresh()]
    )


def hist_monitor(name):
    out, latch = aiger.common._fresh(), aiger.common._fresh()
    return aiger.and_gate([name, 'tmp'], out).loopback(
        {'input': 'tmp', 'output': out, 'latch': latch, 'init': True}
    )


def past_monitor(name):
    out, latch = aiger.common._fresh(), aiger.common._fresh()
    return aiger.or_gate([name, 'tmp'], out).loopback(
        {'input': 'tmp', 'output': out, 'latch': latch, 'init': False}
    )


def since_monitor(left, right):
    tmp = aiger.common._fresh()
    latch = aiger.common._fresh()
    left, right = aiger.atom(left), aiger.atom(right)
    active = aiger.atom(tmp)
    update = active.implies(left | right) & (~active).implies(right)

    circ = update.aig['o', {update.output: tmp}]
    return circ.loopback(
        {'input': tmp, 'output': tmp, 'latch': latch, 'init': False}

    )


def parse(pltl_str: str, output=None):
    expr = PLTLVisitor().visit(PLTL_GRAMMAR.parse(pltl_str))
    aig = expr.aig.evolve(comments=(pltl_str,))
    if output is not None:
        aig = aig['o', {expr.output: output}]

    return type(expr)(aig)



import hypothesis.strategies as st
from hypothesis import given
from hypothesis_cfg import ContextFreeGrammarStrategy

from aiger_ptltl.ptltl import parse

GRAMMAR = {
    'phi': (
        ('Unary', 'phi'),
        ('(', 'phi', 'Binary', 'phi', ')'),
        ('[', 'phi', ' S ', 'phi', ']'),
        ('AP', ),
    ),
    'Unary': (('~', ), ('P',), ('H',), ('Z', )),
    'Interval': (('', ), ('[1, 3]', )),
    'Binary': (
        (' | ', ), (' & ', ), (' -> ', ),
    ),
    'AP': (('ap1', ), ('ap2', ), ('ap3', ), ('ap4', ), ('ap5', )),
}

PTLTL_STRATEGY = st.builds(
    lambda term: parse(''.join(term)),
    ContextFreeGrammarStrategy(GRAMMAR, max_length=14, start='phi')
)


TRUE = {'ap1': True, 'ap2': True, 'ap3': True, 'ap4': True, 'ap5': True}
VALUE_STRATEGY = st.fixed_dictionaries({
    'ap1': st.booleans(),
    'ap2': st.booleans(),
    'ap3': st.booleans(),
    'ap4': st.booleans(),
    'ap5': st.booleans(),
})
TRACE_STRATEGY = st.lists(VALUE_STRATEGY, min_size=1)


@given(PTLTL_STRATEGY, VALUE_STRATEGY)
def test_smoke_value_eval(expr, val):
    res = expr(val)
    assert isinstance(res, bool)


@given(PTLTL_STRATEGY, PTLTL_STRATEGY, TRACE_STRATEGY)
def test_hist_identity(expr1, expr2, trc):
    expr3 = expr1.historically() & expr2.historically()
    expr4 = (expr1 & expr2).historically()
    assert expr3(trc) == expr4(trc)


@given(PTLTL_STRATEGY, PTLTL_STRATEGY, TRACE_STRATEGY)
def test_past_identity(expr1, expr2, trc):
    expr3 = expr1.once() | expr2.once()
    expr4 = (expr1 | expr2).once()
    assert expr3(trc) == expr4(trc)


@given(PTLTL_STRATEGY, TRACE_STRATEGY)
def test_since_to_once_reduction(expr, trc):
    expr2 = expr.once()
    expr3 = parse('TRUE').since(expr)
    assert expr2(trc) == expr3(trc)


@given(PTLTL_STRATEGY, TRACE_STRATEGY)
def test_false_since(expr, trc):
    expr2 = parse('FALSE').since(expr)
    assert expr2(trc) == expr(trc)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiger_ptltl']

package_data = \
{'': ['*']}

install_requires = \
['py-aiger>=4.0.0,<5.0.0']

setup_kwargs = {
    'name': 'py-aiger-ptltl',
    'version': '1.2.0',
    'description': 'Library for generating (p)ast (t)ense (l)inear (t)emporal (l)ogic monitors as aiger circuits.',
    'long_description': "# py-aiger-past-ltl\n\nLibrary for generating (p)ast (t)ense (l)inear (t)emporal (l)ogic\nmonitors as aiger circuits. Builds on the [py-aiger](https://github.com/mvcisback/py-aiger) project.\n\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger-past-ltl/status.svg)](https://cloud.drone.io/mvcisback/py-aiger-past-ltl)\n[![codecov](https://codecov.io/gh/mvcisback/py-aiger-past-ltl/branch/master/graph/badge.svg)](https://codecov.io/gh/mvcisback/py-aiger-past-ltl)\n[![PyPI version](https://badge.fury.io/py/py-aiger-ptltl.svg)](https://badge.fury.io/py/py-aiger-ptltl)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n<!-- markdown-toc start - Don't edit this section. Run M-x markdown-toc-generate-toc again -->\n**Table of Contents**\n\n- [Installation](#installation)\n- [Usage](#usage)\n\n<!-- markdown-toc end -->\n\n\n# Installation\n\nIf you just need to use `aiger_ptltl`, you can just run:\n\n`$ pip install py-aiger-ptltl`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then\nrun:\n\n`$ poetry install`\n\n# Usage\n\nThe primary entry point for using `aiger_ptltl` is the `PTLTLExpr`\nclass which is a simple extension of `aiger.BoolExpr` to support the\ntemporal operators, historically, past (once), (variant) yesterday,\nand since.\n\n```python\nimport aiger_ptltl as ptltl\n\n# Atomic Propositions\nx = ptltl.atom('x')\ny = ptltl.atom('y')\nz = ptltl.atom('z')\n\n# Propositional logic\nexpr1 = ~x\nexpr2 = x & (y | z)\nexpr3 = (x & y) | ~z\nexpr4 = ~(x & y & z)\n\n# Temporal Logic\nexpr5 = x.historically()  #  (H x) ≡ x has held for all previous cycles (inclusive).\nexpr6 = x.once()  #  (P x) ≡ x once held in a past cycle (inclusive).\nexpr7 = x.vyest()  #  (Z x) ≡ x held in the previous cycle (true at time = 0).\nexpr8 = x.since(y)  #  [x S y] ≡ x has held since the cycle after y last held.\n\n# Composition\nexpr9 = expr7.since(expr8.vyest().vyest())\n```\n",
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mvcisback/py-aiger-past-ltl',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
